@@ -20,20 +20,31 @@ Example Usage::
 Thanks go out to Matt Croydon & Christian Metts for putting me up to this late
 at night. The joking around has become reality. :)
 """
+import base64
 import cgi
+import datetime
+import hashlib
+import hmac
+import logging
 import mimetypes
+import numbers
 import os
 import re
 import StringIO
 import sys
+import time
 import traceback
 try:
     from urlparse import parse_qs
 except ImportError:
     from cgi import parse_qs
+try:
+    import Cookie
+except ImportError:
+    import http.cookies as Cookie
 
 __author__ = 'Daniel Lindsley'
-__version__ = ('0', '8', '1')
+__version__ = ('0', '8', '2')
 __license__ = 'BSD'
 
 
@@ -147,6 +158,234 @@ class lazyproperty(object):
         setattr(obj, self._function.func_name, value)
         return value
 
+if hasattr(hmac, 'compare_digest'):  # python 3.3
+    _time_independent_equals = hmac.compare_digest
+else:
+    def _time_independent_equals(a, b):
+        if len(a) != len(b):
+            return False
+        result = 0
+        if isinstance(a[0], int):  # python3 byte strings
+            for x, y in zip(a, b):
+                result |= x ^ y
+        else:  # python2
+            for x, y in zip(a, b):
+                result |= ord(x) ^ ord(y)
+        return result == 0
+
+if type('') is not type(b''):
+    def u(s):
+        return s
+    bytes_type = bytes
+    unicode_type = str
+    basestring_type = str
+else:
+    def u(s):
+        return s.decode('unicode_escape')
+    bytes_type = str
+    unicode_type = unicode
+    basestring_type = basestring
+
+_UTF8_TYPES = (bytes_type, type(None))
+
+
+def utf8(value):
+    """Converts a string argument to a byte string.
+    """
+    if isinstance(value, _UTF8_TYPES):
+        return value
+    assert isinstance(value, unicode_type)
+    return value.encode("utf-8")
+
+_TO_UNICODE_TYPES = (unicode_type, type(None))
+
+
+def to_unicode(value):
+    """Converts a string argument to a unicode string.
+    """
+    if isinstance(value, _TO_UNICODE_TYPES):
+        return value
+    assert isinstance(value, bytes_type)
+    return value.decode("utf-8")
+
+_unicode = to_unicode
+
+if str is unicode_type:
+    native_str = to_unicode
+else:
+    native_str = utf8
+
+
+def format_timestamp(ts):
+    """Formats a timestamp in the format used by HTTP.
+    """
+    if isinstance(ts, (tuple, time.struct_time)):
+        pass
+    elif isinstance(ts, datetime.datetime):
+        ts = ts.utctimetuple()
+    elif isinstance(ts, numbers.Real):
+        ts = time.gmtime(ts)
+    else:
+        raise TypeError("unknown timestamp type: %r" % ts)
+    return time.strftime("%a, %d %b %Y %H:%M:%S GMT", ts)
+
+
+def create_signed_value(secret, name, value):
+    timestamp = utf8(str(int(time.time())))
+    value = base64.b64encode(utf8(value))
+    signature = _create_signature(secret, name, value, timestamp)
+    value = b"|".join([value, timestamp, signature])
+    return value
+
+
+def decode_signed_value(secret, name, value, max_age_days=31):
+    if not value:
+        return None
+    parts = utf8(value).split(b"|")
+    if len(parts) != 3:
+        return None
+    signature = _create_signature(secret, name, parts[0], parts[1])
+    if not _time_independent_equals(parts[2], signature):
+        logging.warning("Invalid cookie signature %r", value)
+        return None
+    timestamp = int(parts[1])
+    if timestamp < time.time() - max_age_days * 86400:
+        logging.warning("Expired cookie %r", value)
+        return None
+    if timestamp > time.time() + 31 * 86400:
+        # _cookie_signature does not hash a delimiter between the
+        # parts of the cookie, so an attacker could transfer trailing
+        # digits from the payload to the timestamp without altering the
+        # signature.  For backwards compatibility, sanity-check timestamp
+        # here instead of modifying _cookie_signature.
+        logging.warning("Cookie timestamp in future; possible tampering %r", value)
+        return None
+    if parts[1].startswith(b"0"):
+        logging.warning("Tampered cookie %r", value)
+        return None
+    try:
+        return base64.b64decode(parts[0])
+    except Exception:
+        return None
+
+
+def _create_signature(secret, *parts):
+    hash = hmac.new(utf8(secret), digestmod=hashlib.sha1)
+    for part in parts:
+        hash.update(utf8(part))
+    return utf8(hash.hexdigest())
+
+
+class HTTPHeaders(dict):
+    """A dictionary that maintains Http-Header-Case for all keys.
+    """
+    def __init__(self, *args, **kwargs):
+        dict.__init__(self)
+        self._as_list = {}
+        self._last_key = None
+        if (len(args) == 1 and len(kwargs) == 0 and
+                isinstance(args[0], HTTPHeaders)):
+            for k, v in args[0].get_all():
+                self.add(k, v)
+        else:
+            self.update(*args, **kwargs)
+
+    def add(self, name, value):
+        """Adds a new value for the given key."""
+        norm_name = HTTPHeaders._normalize_name(name)
+        self._last_key = norm_name
+        if norm_name in self:
+            dict.__setitem__(self, norm_name,
+                             native_str(self[norm_name]) + ',' +
+                             native_str(value))
+            self._as_list[norm_name].append(value)
+        else:
+            self[norm_name] = value
+
+    def get_list(self, name):
+        """Returns all values for the given header as a list."""
+        norm_name = HTTPHeaders._normalize_name(name)
+        return self._as_list.get(norm_name, [])
+
+    def get_all(self):
+        """Returns an iterable of all (name, value) pairs.
+
+        If a header has multiple values, multiple pairs will be
+        returned with the same name.
+        """
+        for name, list in self._as_list.items():
+            for value in list:
+                yield (name, value)
+
+    def parse_line(self, line):
+        """Updates the dictionary with a single header line.
+        """
+        if line[0].isspace():
+            # continuation of a multi-line header
+            new_part = ' ' + line.lstrip()
+            self._as_list[self._last_key][-1] += new_part
+            dict.__setitem__(self, self._last_key,
+                             self[self._last_key] + new_part)
+        else:
+            name, value = line.split(":", 1)
+            self.add(name, value.strip())
+
+    @classmethod
+    def parse(cls, headers):
+        """Returns a dictionary from HTTP header text.
+        """
+        h = cls()
+        for line in headers.splitlines():
+            if line:
+                h.parse_line(line)
+        return h
+
+    def __setitem__(self, name, value):
+        norm_name = HTTPHeaders._normalize_name(name)
+        dict.__setitem__(self, norm_name, value)
+        self._as_list[norm_name] = [value]
+
+    def __getitem__(self, name):
+        return dict.__getitem__(self, HTTPHeaders._normalize_name(name))
+
+    def __delitem__(self, name):
+        norm_name = HTTPHeaders._normalize_name(name)
+        dict.__delitem__(self, norm_name)
+        del self._as_list[norm_name]
+
+    def __contains__(self, name):
+        norm_name = HTTPHeaders._normalize_name(name)
+        return dict.__contains__(self, norm_name)
+
+    def get(self, name, default=None):
+        return dict.get(self, HTTPHeaders._normalize_name(name), default)
+
+    def update(self, *args, **kwargs):
+        for k, v in dict(*args, **kwargs).items():
+            self[k] = v
+
+    def copy(self):
+        return HTTPHeaders(self)
+
+    _NORMALIZED_HEADER_RE = re.compile(
+        r'^[A-Z0-9][a-z0-9]*(-[A-Z0-9][a-z0-9]*)*$')
+    _normalized_headers = {}
+
+    @staticmethod
+    def _normalize_name(name):
+        """Converts a name to Http-Header-Case.
+        """
+        try:
+            return HTTPHeaders._normalized_headers[name]
+        except KeyError:
+            if HTTPHeaders._NORMALIZED_HEADER_RE.match(name):
+                normalized = name
+            else:
+                normalized = "-".join(
+                    [w.capitalize() for w in name.split("-")])
+            HTTPHeaders._normalized_headers[name] = normalized
+            return normalized
+
 
 class Request(object):
     """An object to wrap the environ bits in a friendlier way."""
@@ -162,7 +401,14 @@ class Request(object):
         self.method = self._environ.get('REQUEST_METHOD', 'GET').upper()
         self.query = self._environ.get('QUERY_STRING', '')
         self.content_length = 0
-
+        self.headers = HTTPHeaders()
+        if self._environ.get("CONTENT_TYPE"):
+            self.headers["Content-Type"] = self._environ["CONTENT_TYPE"]
+        if self._environ.get("CONTENT_LENGTH"):
+            self.headers["Content-Length"] = self._environ["CONTENT_LENGTH"]
+        for key in self._environ:
+            if key.startswith("HTTP_"):
+                self.headers[key[5:].replace("_", "-")] = self._environ[key]
         try:
             self.content_length = int(self._environ.get('CONTENT_LENGTH', '0'))
         except ValueError:
@@ -192,6 +438,33 @@ class Request(object):
         """Content of the request."""
         return self._environ['wsgi.input'].read(self.content_length)
 
+    @property
+    def cookies(self):
+        """A dictionary of Cookie.Morsel objects."""
+        if not hasattr(self, "_cookies"):
+            self._cookies = Cookie.SimpleCookie()
+            if "Cookie" in self.headers:
+                try:
+                    self._cookies.load(
+                        native_str(self.headers["Cookie"]))
+                except Exception:
+                    self._cookies = None
+        return self._cookies
+
+    def get_cookie(self, name, default=None):
+        """Gets the value of the cookie with the given name, else default."""
+        if self.cookies is not None and name in self.cookies:
+            return self.cookies[name].value
+        return default
+
+    def get_secure_cookie(self, name, value=None, max_age_days=31):
+        """Returns the given signed cookie if it validates, or None.
+        """
+        if value is None:
+            value = self.get_cookie(name)
+        return decode_signed_value(COOKIE_SECRET, name, value,
+                                   max_age_days=max_age_days)
+
     def build_get_dict(self):
         """Takes GET data and rips it apart into a dict."""
         raw_query_dict = parse_qs(self.query, keep_blank_values=1)
@@ -227,30 +500,81 @@ class Request(object):
 
 
 class Response(object):
-    headers = []
 
     def __init__(self, output, headers=None, status=200, content_type='text/html'):
         self.output = output
         self.content_type = content_type
         self.status = status
-        self.headers = []
+        self.headers = HTTPHeaders()
 
-        if headers and isinstance(headers, list):
+        if headers and isinstance(headers, HTTPHeaders):
             self.headers = headers
+        if headers and isinstance(headers, list):
+            for (key, value) in headers:
+                self.headers.add(key, value)
 
     def add_header(self, key, value):
-        self.headers.append((key, value))
+        self.headers.add(key, value)
+
+    def set_cookie(self, name, value, domain=None, expires=None, path="/",
+                   expires_days=None, **kwargs):
+        """Sets the given cookie name/value with the given options.
+        """
+        name = native_str(name)
+        value = native_str(value)
+        if re.search(r"[\x00-\x20]", name + value):
+            raise ValueError("Invalid cookie %r: %r" % (name, value))
+        if not hasattr(self, "_new_cookie"):
+            self._new_cookie = Cookie.SimpleCookie()
+        if name in self._new_cookie:
+            del self._new_cookie[name]
+        self._new_cookie[name] = value
+        morsel = self._new_cookie[name]
+        if domain:
+            morsel["domain"] = domain
+        if expires_days is not None and not expires:
+            expires = datetime.datetime.utcnow() + datetime.timedelta(
+                days=expires_days)
+        if expires:
+            morsel["expires"] = format_timestamp(expires)
+        if path:
+            morsel["path"] = path
+        for k, v in kwargs.items():
+            if k == 'max_age':
+                k = 'max-age'
+            morsel[k] = v
+
+    def clear_cookie(self, name, path="/", domain=None):
+        """Deletes the cookie with the given name."""
+        expires = datetime.datetime.utcnow() - datetime.timedelta(days=365)
+        self.set_cookie(name, value="", path=path, expires=expires,
+                        domain=domain)
+
+    def clear_all_cookies(self):
+        """Deletes all the cookies the user sent with this request."""
+        for name in self.request.cookies:
+            self.clear_cookie(name)
+
+    def set_secure_cookie(self, name, value, expires_days=30, **kwargs):
+        """Signs and timestamps a cookie so it cannot be forged."""
+        self.set_cookie(name, self.create_signed_value(name, value),
+                        expires_days=expires_days, **kwargs)
+
+    def create_signed_value(self, name, value):
+        """Signs and timestamps a string so it cannot be forged.
+        """
+        return create_signed_value(COOKIE_SECRET, name, value)
 
     def send(self, start_response):
         status = "%d %s" % (self.status, HTTP_MAPPINGS.get(self.status))
-        headers = [('Content-Type', "%s; charset=utf-8" % self.content_type)] + self.headers
-        final_headers = []
+        headers = ([('Content-Type', "%s; charset=utf-8" % self.content_type)] +
+                  [(k, v) for k, v in self.headers.iteritems()])
 
-        # Because Unicode is unsupported...
-        for header in headers:
-            final_headers.append((self.convert_to_ascii(header[0]), self.convert_to_ascii(header[1])))
+        if hasattr(self, "_new_cookie"):
+            for cookie in self._new_cookie.values():
+                headers.append(("Set-Cookie", utf8(cookie.OutputString(None))))
 
-        start_response(status, final_headers)
+        start_response(status, headers)
 
         if isinstance(self.output, unicode):
             return self.output.encode('utf-8')
@@ -592,9 +916,13 @@ WSGI_ADAPTERS = {
 }
 
 
+COOKIE_SECRET = None
+
 # Server
 
-def run_itty(server='wsgiref', host='localhost', port=8080, config=None):
+
+def run_itty(server='wsgiref', host='localhost', port=8080, config=None,
+    cookie_secret=None):
     """
     Runs the itty web server.
 
@@ -621,6 +949,9 @@ def run_itty(server='wsgiref', host='localhost', port=8080, config=None):
         print 'Use Ctrl-C to quit.'
         print
 
+    global COOKIE_SECRET
+    COOKIE_SECRET = cookie_secret or base64.b64encode(os.urandom(32))
+
     try:
         WSGI_ADAPTERS[server](host, port)
     except KeyboardInterrupt:
